import boto3
import os
import json
import pandas as pd
import numpy as np
from io import BytesIO, StringIO
import pickle
import h5py
import requests
from concurrent.futures import ThreadPoolExecutor
from datetime import date
import time

class DataExtractor():
    """
    Clase para extraer datos históricos de precios utilizando la API de EODHD.
    
    Atributos:
    ----------
    eod_token : str
        Token de acceso para la API de EODHD.
    """

    def __init__(self, eod_token):
        """
        Inicializa la clase con el token de la API.

        Parámetros:
        -----------
        eod_token : str
            Token para la autenticación en la API de EODHD.
        """
        self.eod_token = eod_token
    def eod_extract(self, symbol, freq, start_date=None, end_date=None, exchange='US', fmt='csv', intraday_spot=['Close'], historical_spot=['Adjusted_close']):
        """
        Extrae datos históricos o intradía de un símbolo financiero utilizando la API de EOD Historical Data.
    
        Parámetros:
        -----------
        symbol : str
            El ticker del activo financiero (e.g., 'AAPL' para Apple).
        freq : str
            La frecuencia de los datos a extraer. Valores permitidos: ['m', 'w', 'd', '1h', '5m', '1m'].
            'm' - mensual, 'w' - semanal, 'd' - diario, '1h' - cada hora, '5m' - cada 5 minutos, '1m' - cada minuto.
        start_date : int, opcional
            Timestamp de inicio para la extracción de datos en formato UNIX. Si no se especifica, calcula automáticamente
            basado en la máxima duración permitida por la frecuencia elegida.
        end_date : int, opcional
            Timestamp de fin para la extracción de datos en formato UNIX. Si no se especifica, se usa el timestamp actual.
        exchange : str, opcional
            El código del mercado de valores en el que se negocia el símbolo, por defecto es 'US'.
        fmt : str, opcional
            El formato de salida de los datos. Valores permitidos: 'csv' (por defecto), 'json'.
        intraday_spot : list, opcional
            Las columnas específicas a retornar para datos intradía. Por defecto es ['Close'].
        historical_spot : list, opcional
            Las columnas específicas a retornar para datos históricos. Por defecto es ['Adjusted_close'].
    
        Retorna:
        --------
        pd.DataFrame
            Un DataFrame de pandas con las columnas especificadas, indexado por la fecha y hora de los datos.
            Para datos intradía, el índice será 'Datetime' y para datos históricos, será 'Date'.
    
        Levanta:
        --------
        ValueError
            Si la 'freq' no está en las frecuencias válidas o si los periodos entre 'start_date' y 'end_date'
            exceden los máximos permitidos para la frecuencia especificada.
    
        Ejemplo:
        --------
        >>> data_extractor = DataExtractor(eod_token='your_api_token')
        >>> aapl_data = data_extractor.eod_extract_('AAPL', '1h', exchange='US', fmt='csv')
        >>> print(aapl_data.head())
        """
        valid_frequencies = ['m', 'w', 'd', '1h', '5m', '1m']
        if freq not in valid_frequencies:
            raise ValueError(f"freq parameter must be one of: {valid_frequencies}")
        if freq in ['1h', '5m', '1m']:
            if not end_date:
                end_date = int(time.time())
            if not start_date:
                start_date = end_date - 120 * 24 * 60 * 60
            if ((end_date - start_date) / (24 * 3600) > 120) & (freq == '1m'):
                raise ValueError('maximum periods between dates are 120 days for 1-minute frequency')
            elif ((end_date - start_date) / (24 * 3600) > 600) & (freq == '5m'):
                raise ValueError('maximum periods between dates are 600 days for 5-minute frequency')
            elif ((end_date - start_date) / (24 * 3600) > 7200) & (freq == '1h'):
                raise ValueError('maximum periods between dates are 7200 days for 1-hour frequency')
            url = f'https://eodhd.com/api/intraday/{symbol}.{exchange}?from={start_date}&to={end_date}&interval={freq}&api_token={self.eod_token}&fmt={fmt}'
            data = requests.get(url).content
            df = self._decode_(data)
            df['Datetime'] = pd.to_datetime(df['Datetime'])
            df.set_index('Datetime', inplace=True)
            return df[intraday_spot]
        else:
            if not end_date:
                end_date = self._today_date_()
            if not start_date:
                start_date = '1900-01-01'
            url = f'https://eodhd.com/api/eod/{symbol}.{exchange}?from={start_date}&to={end_date}&period={freq}&api_token={self.eod_token}&fmt={fmt}'
            data = requests.get(url).content
            df = self._decode_(data)
            df['Date'] = pd.to_datetime(df['Date'])
            df.set_index('Date', inplace=True)
            return df[historical_spot]

    def extract_multiple(self, symbols, freq='d', start_date=None, end_date=None, exchange='US', fmt='csv', intraday_spot=['Close'], historical_spot=['Adjusted_close'], max_workers=5):
        """
        Extrae datos históricos o intradía para múltiples símbolos de forma paralela.
    
        Parámetros:
        -----------
        symbols : list
            Lista de símbolos de los activos financieros.
        freq : str, opcional
            Frecuencia de los datos ('m', 'w', 'd', '1h', '5m', '1m'). Por defecto es 'd' para datos diarios.
        start_date : int, opcional
            Timestamp de inicio para la extracción de datos en formato UNIX. Si no se especifica, se calcula automáticamente.
        end_date : int, opcional
            Timestamp de fin para la extracción de datos en formato UNIX. Si no se especifica, se utiliza el timestamp actual.
        exchange : str, opcional
            Código del mercado de valores. Por defecto es 'US'.
        fmt : str, opcional
            Formato de los datos recibidos (‘json’ o ‘csv’). Por defecto es 'csv'.
        intraday_spot : list, opcional
            Columnas que se devolverán para datos intradía. Por defecto es ['Close'].
        historical_spot : list, opcional
            Columnas que se devolverán para datos históricos. Por defecto es ['Adjusted_close'].
        max_workers : int, opcional
            Número máximo de hilos de ejecución en paralelo. Por defecto es 5.
    
        Retorna:
        --------
        pd.DataFrame
            DataFrame con los datos históricos o intradía de los símbolos, con las columnas nombradas directamente por los tickers.
        """
    
        def task(symbol):
            return symbol, self.eod_extract(symbol, freq, start_date, end_date, exchange, fmt, intraday_spot, historical_spot)
    
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            results = list(executor.map(task, symbols))

        combined_df = pd.concat([df for _, df in results], axis=1)
        combined_df.columns = symbols
    
        return combined_df

    def _decode_(self, data):
        """
        Decodifica el contenido recibido en formato CSV y lo convierte en un DataFrame.

        Parámetros:
        -----------
        data : bytes
            Datos en formato binario (CSV codificado en bytes).

        Retorna:
        --------
        pd.DataFrame
            DataFrame con los datos decodificados.
        """
        df = pd.read_csv(StringIO(data.decode('utf-8')))
        return df

    def _today_date_(self):
        """
        Obtiene la fecha de hoy en formato 'YYYY-MM-DD'.

        Retorna:
        --------
        str
            Fecha actual en formato 'YYYY-MM-DD'.
        """
        return date.today().strftime("%Y-%m-%d")

class S3Touch():
    """
    Clase S3Touch para interactuar con Amazon S3. Permite subir archivos o directorios completos a un bucket de S3
    y leer archivos desde S3 en diferentes formatos como JSON, CSV y Numpy (.npy).

    Atributos:
    ----------
    bucket_name : str
        El nombre del bucket de S3 con el cual se desea interactuar.
    
    s3 : boto3.client
        El cliente de boto3 que permite realizar operaciones con S3.

    Métodos:
    --------
    __init__(bucket_name, access_key, secret_access_key, region_name):
        Inicializa la clase con las credenciales y la configuración para interactuar con S3.
    
    _upload_file(file_path, s3_folder=None):
        Sube un archivo único desde el sistema local a un bucket de S3.
    
    _upload_folder(folder_path, s3_folder=None):
        Sube un directorio completo de archivos desde el sistema local a un bucket de S3.
    
    write(path, s3_folder=None):
        Detecta si el path es un archivo o directorio y llama a los métodos correspondientes para subirlos a S3.
    
    read(s3_key, local_path=None):
        Lee un archivo de S3 y lo procesa en base a su extensión (JSON, CSV, Numpy).
    """
    
    def __init__(self, bucket_name, access_key, secret_access_key, region_name):
        """
        Inicializa la clase S3Touch con las credenciales y configuraciones necesarias para interactuar con S3.
        
        Parámetros:
        -----------
        bucket_name : str
            El nombre del bucket de S3 al que se subirán o leerán archivos.
        
        access_key : str
            Clave de acceso (AWS Access Key) para autenticar el cliente S3.
        
        secret_access_key : str
            Clave secreta (AWS Secret Access Key) asociada al acceso S3.
        
        region_name : str
            Región de AWS en la que se encuentra el bucket de S3.
        """
        self.bucket_name = bucket_name
        self.s3 = boto3.client('s3',
                               aws_access_key_id=access_key,
                               aws_secret_access_key=secret_access_key,
                               region_name=region_name)

    def _upload_file(self, file_path, s3_folder=None):
        """
        Sube un archivo único al bucket de S3.

        Parámetros:
        -----------
        file_path : str
            Ruta completa del archivo en el sistema local que se va a subir.
        
        s3_folder : str, opcional
            Carpeta dentro del bucket de S3 donde se guardará el archivo. Si no se especifica, se sube a la raíz del bucket.
        
        Excepciones:
        ------------
        Podría generar excepciones relacionadas con permisos o accesos a S3 si el cliente no tiene acceso adecuado.
        """
        s3_key = f"{s3_folder}/{os.path.basename(file_path)}" if s3_folder else os.path.basename(file_path)
        s3_key = s3_key.lstrip('/')  # Elimina barra inicial si existe
        self.s3.upload_file(file_path, self.bucket_name, s3_key)

    def _upload_folder(self, folder_path, s3_folder=None):
        """
        Sube todos los archivos dentro de un directorio al bucket de S3, manteniendo la estructura de subcarpetas.

        Parámetros:
        -----------
        folder_path : str
            Ruta del directorio en el sistema local que se va a subir.
        
        s3_folder : str, opcional
            Carpeta dentro del bucket de S3 donde se guardará la estructura de archivos. Si no se especifica, se suben a la raíz del bucket.
        
        Excepciones:
        ------------
        Podría generar excepciones si no tiene acceso a S3 o si hay problemas con los permisos.
        """
        if not s3_folder:
            s3_folder = f'{os.path.basename(folder_path)}/'
        else:
            s3_folder = f'{s3_folder}/'
        for root, dirs, files in os.walk(folder_path):
            for file in files:
                file_path = os.path.join(root, file)
                relative_path = os.path.relpath(file_path, folder_path)
                s3_key = os.path.join(s3_folder, relative_path).replace("\\", "/")
                self.s3.upload_file(file_path, self.bucket_name, s3_key)
                print (f'{file} succesfully wrote')

    def write(self, path, s3_folder=None):
        """
        Sube un archivo o directorio al bucket de S3. Si la ruta proporcionada es un archivo, lo sube. Si es un directorio, sube todo su contenido.

        Parámetros:
        -----------
        path : str
            Ruta del archivo o directorio que se desea subir.
        
        s3_folder : str, opcional
            Carpeta dentro del bucket de S3 donde se subirá el archivo o directorio. Si no se especifica, se sube a la raíz del bucket.
        
        Excepciones:
        ------------
        ValueError: Si la ruta proporcionada no es ni un archivo ni un directorio.
        """
        if os.path.isfile(path):
            self._upload_file(path, s3_folder)
        elif os.path.isdir(path):
            self._upload_folder(path, s3_folder)
        else:
            raise ValueError("La ruta proporcionada no es válida. Debe ser un archivo o un directorio.")
            
    def read(self, s3_key, local_path=None):
        """
        Lee un archivo desde el bucket de S3 y lo procesa según su tipo de archivo.
    
        Parámetros:
        -----------
        s3_key : str
            Clave (ruta) del archivo dentro del bucket de S3.
        
        local_path : str, opcional
            Ruta local donde se desea descargar el archivo. Si se especifica, el archivo se guardará localmente en lugar de solo procesarlo.
        
        Retorno:
        --------
        El archivo procesado, que puede ser:
        - Un diccionario (para archivos JSON).
        - Un array de Numpy (para archivos .npy).
        - Un DataFrame de pandas (para archivos .csv).
        - Un objeto Python (para archivos .pkl).
        - Un objeto h5py.File (para archivos .h5).
        
        Excepciones:
        ------------
        ValueError: Si el formato del archivo no es soportado.
        """
        try:
            response = self.s3.get_object(Bucket=self.bucket_name, Key=s3_key)
            file_content = response['Body'].read()
    
            if s3_key.endswith('.json'):
                file = json.loads(file_content.decode('utf-8'))
            elif s3_key.endswith('.npy'):
                file = np.load(BytesIO(file_content), allow_pickle=True)
            elif s3_key.endswith('.csv'):
                file = pd.read_csv(StringIO(file_content.decode('utf-8')))
            elif s3_key.endswith('.pkl'):
                file = pickle.loads(file_content)
            elif s3_key.endswith('.h5'):
                # Crear un archivo temporal para poder abrirlo con h5py
                with open('temp_model.h5', 'wb') as f:
                    f.write(file_content)
                file = h5py.File('temp_model.h5', 'r')
            else:
                raise ValueError(f"Formato de archivo no soportado: {s3_key}")
    
            if local_path:
                with open(local_path, 'wb') as f:
                    f.write(file_content)
            
            return file
    
        except Exception as e:
            print(f"Error al procesar el archivo: {e}")
            raise e
    def download_folder (self, s3_folder, local_path):
        """
        Download an entire folder from an S3 bucket to a local directory.
    
        This method downloads all the files located in a specified folder in S3 to a local directory. 
        It checks if the folder path on S3 ends with a '/' and adjusts it if necessary. It then lists all
        objects in the specified S3 folder. If the folder contains any files (not just subdirectories),
        each file is downloaded to the corresponding local path.
    
        Parameters:
        - s3_folder (str): The folder path in the S3 bucket. Must not be an empty string.
        - local_path (str): The local directory path where the files will be downloaded.
    
        Notes:
        - This function assumes that 'self.s3' is an initialized boto3 S3 client and 'self.bucket_name'
          is the name of the S3 bucket.
        - The function creates any necessary local directories if they do not exist.
        - Files are saved in the local directory maintaining their relative path as in S3.
    
        """
        if not s3_folder.endswith('/'):
            s3_folder += '/'
        response = self.s3.list_objects_v2(Bucket=self.bucket_name, Prefix=s3_folder)
        if 'Contents' in response:
            for file in response['Contents']:
                file_name = file['Key']
                if not file_name.endswith('/'):
                    local_file_path = os.path.join(local_path, os.path.relpath(file_name, s3_folder))
                    os.makedirs(os.path.dirname(local_file_path), exist_ok=True)
                    s3.download_file(bucket_name, file_name, local_file_path)
                    print(f"Downloaded {file_name} to {local_file_path}")

def invoke_lambda_to_stop_instance(access_key, secret_access_key, region_name, instance_id):
    """
    Invoca una función Lambda de AWS para detener una instancia EC2 específica.

    Esta función utiliza las credenciales de AWS y los parámetros proporcionados para invocar 
    una función Lambda que está configurada para detener una instancia EC2. La invocación 
    se realiza de manera asíncrona (InvocationType='Event').

    Parámetros:
    -----------
    access_key : str
        Clave de acceso de AWS (AWS access key) necesaria para autenticarse en los servicios de AWS.
    
    secret_access_key : str
        Clave secreta de AWS (AWS secret access key) asociada al acceso.
    
    region_name : str
        Nombre de la región de AWS en la que se encuentra la función Lambda y la instancia EC2.
    
    instance_id : str
        ID de la instancia EC2 que se desea detener mediante la función Lambda.

    Retorna:
    --------
    response : dict
        Respuesta del cliente de Lambda de AWS, que contiene información sobre el resultado de la invocación.
    """
    lambda_client = boto3.client(
        'lambda',
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_access_key,
        region_name=region_name
    )

    payload = {
        'instance_id': instance_id
    }

    response = lambda_client.invoke(
        FunctionName='stop_instance_after_executed',  # Nombre de la función Lambda que detiene la instancia.
        InvocationType='Event',  # 'Event' significa que la invocación es asíncrona.
        Payload=json.dumps(payload)  # Cargar el ID de la instancia como un JSON.
    )
    print ('lambda invoked')
    return response




import os
import chardet
import pandas as pd
import numpy as np
from datetime import datetime
import json

def create_event (event_path, prefix):
    """
    Crea un evento con un nombre basado en un prefijo y un timestamp actual, 
    y genera una carpeta asociada al evento.

    Args:
        event_path (str): La ruta base donde se creará la carpeta del evento.
        prefix (str): El prefijo que se utilizará para nombrar el evento.

    Returns:
        tuple: Una tupla que contiene:
            - event (str): El nombre del evento generado, que consiste en el prefijo
              seguido de un timestamp con el formato 'YYYYMMDD_HHMMSS'.
            - event_folder_path (str): La ruta completa de la carpeta creada para el evento.
    """
    timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
    event = f"{prefix}_{timestamp}"
    event_folder_path = crear_carpeta(event_path, event)
    return event, event_folder_path

def crear_carpeta(ruta_base, nombre_carpeta):
    """
    Crea una carpeta dada la ruta base y el nombre deseado

    Args:
        ruta_base(str): Ruta de la carpeta raíz
        nombre_carpeta(str): Nombre deseado para la carpeta
    Returns:
        str: Ruta completa de la carpeta creada
    """
    try:
        ruta_completa = os.path.join(ruta_base, nombre_carpeta)
        os.makedirs(ruta_completa, exist_ok=True)
        print(f"Carpeta creada en: {ruta_completa}")
        return ruta_completa
    except Exception as e:
        print(f"Ocurrió un error al crear la carpeta: {e}")

def save_array(path, file_name, array):
    """
    Guarda un array en formato .npy en la ruta especificada.

    Parámetros:
    path (str): La ruta completa, incluyendo el nombre del archivo donde se guardará el array.
    file_name (str): Nombre con el que se desea guardar el archivo
    array (numpy.ndarray): El array que se desea guardar.

    Ejemplo:
    save_array('ruta/al/archivo.npy', mi_array)
    """
    try:
        file_path = os.path.join(path, f'{file_name}.npy')
        np.save(file_path, array)
        print(f"Array guardado exitosamente en: {file_path}")
    except Exception as e:
        print(f"Error al guardar el array: {e}")

def read_codificated_csv(path):
    """
    Detecta la codificación del archivo .csv y lo lee utilizando esa codificación
    Args:
        path(str): Ruta al archivo .csv
    Returns:
        df(pd.DataFrame): Dataframe leído
    """
    with open(fr'{path}', 'rb') as f:
        result = chardet.detect(f.read())
    return pd.read_csv(fr'{path}', encoding=result['encoding'])


from setuptools import setup, find_packages

setup(
    name='godolib',
    version='1.1.8',
    packages=find_packages(),
    install_requires=['statsmodels', 'numpy', 'scikit-learn', 'pandas', 'boto3', 'chardet', 'psutil', 'plotly', 'matplotlib', 'requests', 'h5py', 'tensorflow'],
    description='Machine/Deep learning and preprocessing oriented library',
    author='Sergio Montes',
    author_email='ss.montes.jimenez@gmail.com',
    url='',
)



import psutil
import os

def memory_usage():
    """
    Imprime la cantidad de memoria RAM utilizada por el proceso actual en megabytes (MB).

    Esta función utiliza la biblioteca `psutil` para acceder a la información del sistema, 
    específicamente al uso de memoria del proceso que está ejecutando el código. La memoria 
    utilizada se mide en bytes y se convierte a megabytes para una mejor legibilidad.

    Ejemplo:
        >>> obtener_uso_memoria()
        Uso de memoria: 120.35 MB
    """
    proceso = psutil.Process(os.getpid())
    memoria_en_mb = proceso.memory_info().rss / (1024 * 1024)
    print(f"Used Memory: {memoria_en_mb:.2f} MB")


import tensorflow as tf

def r2(y_true, y_pred):
    """
    Calcula el coeficiente de determinación R² para un modelo de regresión. Este coeficiente, también conocido como
    R cuadrado, proporciona una medida de cuán bien los resultados observados son replicados por el modelo, basado
    en la proporción de la variación total de los datos explicada por el modelo.

    Parámetros:
    -----------
    y_true : Tensor
        Tensor que contiene los valores verdaderos. Debe ser de la misma forma que y_pred.
    y_pred : Tensor
        Tensor que contiene los valores predichos por el modelo de machine learning.

    Retorna:
    --------
    r2_score : Tensor
        Un tensor escalar que contiene el valor del coeficiente de determinación R². Un valor de R² de 1 indica
        que el modelo explica perfectamente la variabilidad de los datos de respuesta alrededor de su media,
        mientras que un valor de 0 indica que el modelo no explica nada de la variabilidad.

    Descripción de la función:
    --------------------------
    - ss_res : Suma de cuadrados de los residuos. Se calcula como la suma de los cuadrados de las diferencias
      entre los valores reales y los predichos.
    - ss_tot : Suma total de cuadrados. Es la suma de los cuadrados de las diferencias entre los valores reales
      y la media de esos valores.
    - r2_scores : Tensor que contiene el coeficiente R² calculado para cada característica individualmente,
      ajustado para múltiples dimensiones de salida.
    - r2_score : Valor escalar del R², obtenido como la media de los R² de todas las características, lo cual
      proporciona una única métrica representativa del rendimiento del modelo a lo largo de todas las características
      predichas.

    Nota:
    -----
    La función utiliza `tf.keras.backend.epsilon()` para añadir un pequeño número al denominador en el cálculo
    de R² para evitar divisiones por cero y asegurar la estabilidad numérica del cálculo.
    """
    # Calcula la suma de cuadrados de los residuos
    ss_res = tf.reduce_sum(tf.square(y_true - y_pred), axis=0)
    # Calcula la suma total de cuadrados
    ss_tot = tf.reduce_sum(tf.square(y_true - tf.reduce_mean(y_true, axis=0)), axis=0)
    # Calcula R² para cada salida y promedia sobre todas las salidas
    r2_scores = 1 - ss_res / (ss_tot + tf.keras.backend.epsilon())
    r2_score = tf.reduce_mean(r2_scores)
    
    return r2_score


import random
import os
import pandas as pd
import numpy as np
from .monitors import ModelManager


class ModelTuner ():
    """
    A class to tune machine learning models over multiple hyperparameter configurations and trials.

    This class facilitates the process of hyperparameter tuning by testing multiple configurations of 
    hyperparameters on a provided model architecture over a series of trials. Each configuration is evaluated
    with a rolling window over the data, and the results are cleaned and returned as structured data.

    Attributes:
    -----------
    post_transformation_pipeline : object
        A transformation pipeline to be applied after the main transformations.
        
    transformation_pipeline : object
        A transformation pipeline used to preprocess the original data before model training.
        
    transformed_df : pd.DataFrame
        The transformed DataFrame used for training the models.
        
    recovered_df : pd.DataFrame
        The original DataFrame recovered by applying the inverse transformation to `transformed_df`.
        
    building_func : callable
        A function that builds and returns a machine learning model based on hyperparameters.
        
    directory : str
        Directory path where model weights will be stored after training.

    Methods:
    --------
    search(hyperparameters, X, Y, epochs, num_trials, roll_size, callbacks, verbose=0, batch_size=32, data_freq='D', state='transformed'):
        Performs hyperparameter tuning by iterating over different hyperparameter configurations and trials.
        
    _construct_model_(hps_configuration, X, Y, epochs, batch_size, callbacks, verbose, weights_path=False):
        Builds and trains a model based on a hyperparameter configuration and optionally loads pre-trained weights.
        
    _is_iterable_(obj):
        Helper method to check if the provided object is iterable.
        
    _clean_evaluation_(evaluation):
        Cleans and structures the model evaluation results into a dictionary of dataframes where each key is 
        a metric and each dataframe contains trial results for that metric.
    """
    def __init__ (self, post_transformation_pipeline, transformation_pipeline, transformed_df, building_func, directory):
        """
        Initializes the ModelTuner object with transformation pipelines, data, and model-building function.
        
        Parameters:
        -----------
        post_transformation_pipeline : object
            The pipeline that will be applied to the data after the model predictions.
        
        transformation_pipeline : object
            The pipeline used to transform the input data before training.
        
        transformed_df : pd.DataFrame
            The DataFrame containing the transformed data.
        
        building_func : callable
            A function that takes a hyperparameters object and returns a model.
        
        directory : str
            Path where model weights and logs will be stored.
        """
        self.post_transformation_pipeline=post_transformation_pipeline
        self.transformation_pipeline=transformation_pipeline
        self.transformed_df=transformed_df
        self.recovered_df=transformation_pipeline.inverse_transform(transformed_df)
        self.building_func=building_func
        self.directory=directory
    def search (self, hyperparameters, X, Y, validation_data, epochs, num_trials, roll_size, callbacks, verbose=0, batch_size=32, data_freq='D', state='transformed'):
        """
        Searches over a set of hyperparameter configurations and evaluates models across multiple trials.
        
        Parameters:
        -----------
        hyperparameters : iterable
            A collection of hyperparameter configurations to be tested.
            
        X : np.array or pd.DataFrame
            Input features for training the model.
            
        Y : np.array or pd.DataFrame
            Target values for training the model.

        validation_vata : tuple
            Validation_data
            
        epochs : int
            Number of epochs to train each model.
            
        num_trials : int
            Number of trials for each hyperparameter configuration.
            
        roll_size : int
            Size of the rolling window to use when evaluating model performance.
            
        callbacks : list
            List of callback functions to be used during model training.
            
        verbose : int, optional (default=0)
            Verbosity mode for model training.
            
        batch_size : int, optional (default=32)
            Batch size to use during model training.
            
        data_freq : str, optional (default='D')
            Frequency of the data used for the model.
            
        state : str, optional (default='transformed')
            Indicates whether the state of the data is 'transformed' or 'raw'.
            
        Returns:
        --------
        dict
            A dictionary containing the cleaned evaluation metrics for each hyperparameter configuration and trial.
        """
        if not isinstance(hyperparameters, list):
            raise ValueError(f'hyperparameters parameter must be a list')
        if not all (isinstance(element, dict) for element in hyperparameters):
            raise ValueError(f'hyperparametrs parameter must contain only dictionaries')
        evaluations = {}
        os.makedirs(self.directory, exist_ok=True)
        for hps_idx, hps_configuration in enumerate(hyperparameters):
            print (f'searching configuration: {hps_idx}')
            model_evaluations = {}
            for trial in range (num_trials):
                print (f'trial: {trial}')
                trial_weights_path = f'{self.directory}/{hps_idx}_{trial}.weights.h5'
                if os.path.exists(trial_weights_path):
                    model = self._construct_model_(hps_configuration=hps_configuration,
                                                   X=X,
                                                   Y=Y,
                                                   validation_data=validation_data,
                                                   epochs=epochs,
                                                   batch_size=batch_size,
                                                   callbacks=callbacks,
                                                   verbose=verbose,
                                                   weights_path=trial_weights_path
                                                   )
                else:
                    model = self._construct_model_(hps_configuration=hps_configuration,
                                                   X=X,
                                                   Y=Y,
                                                   validation_data=validation_data,
                                                   epochs=epochs,
                                                   batch_size=batch_size,
                                                   callbacks=callbacks,
                                                   verbose=verbose
                                                   )
                    model.save_weights(filepath=trial_weights_path)
                manager = ModelManager(post_transformation_pipeline=self.post_transformation_pipeline,
                                       transformation_pipeline=self.transformation_pipeline,
                                       transformed_df=self.transformed_df, model=model)
                evaluation = manager.evaluate(roll_size=roll_size, state=state).iloc[:-1, :]
                model_evaluations[trial] = evaluation
            evaluations[hps_idx] = model_evaluations
        cleaned_evaluations = self._clean_evaluation_(evaluation=evaluations)
        return cleaned_evaluations
    def _construct_model_(self, hps_configuration, X, Y, validation_data, epochs, batch_size, callbacks, verbose, weights_path=False):
        """
        Constructs and trains a model based on the provided hyperparameter configuration.
        
        Parameters:
        -----------
        hps_configuration : dict
            Hyperparameter configuration for the model.
            
        X : np.array or pd.DataFrame
            Input features for training.
            
        Y : np.array or pd.DataFrame
            Target values for training.
            
        validation_vata : tuple
            Validation_data
            
        epochs : int
            Number of epochs to train the model.
            
        batch_size : int
            Batch size to use during training.
            
        callbacks : list
            List of callback functions to be used during training.
            
        verbose : int
            Verbosity mode for model training.
            
        weights_path : str, optional (default=False)
            Path to pre-trained weights. If provided, the model will load these weights before training.
            
        Returns:
        --------
        model
            The constructed and trained machine learning model.
        """
        hp_object = HyperparametersSelector(hps_configuration = hps_configuration)
        model = self.building_func(hp_object)
        if weights_path:
            model.load_weights(filepath=weights_path)
            return model
        model.fit(x=X, y=Y, validation_data=validation_data, epochs=epochs, batch_size=batch_size, callbacks=callbacks, verbose=verbose)
        return model
    def _clean_evaluation_ (self, evaluation):
        """
        Cleans and organizes evaluation results into a structured format.
        
        Parameters:
        -----------
        evaluation : dict
            The raw evaluation data, where keys are hyperparameter configurations and values are dictionaries
            of trial results.
            
        Returns:
        --------
        dict
            A dictionary of DataFrames where each key is a metric and each DataFrame contains results for 
            that metric across different hyperparameter configurations and trials.
        """
        metrics_dict = {}
        for hps_config, trials in evaluation.items():
            for trial, df in trials.items():
                for metric in df.columns:
                    if metric not in metrics_dict:
                        metrics_dict[metric] = pd.DataFrame(columns=['Hps_config', 'trial'] + df.index.tolist())
                    new_row = pd.Series([hps_config, trial] + df[metric].tolist(), index=metrics_dict[metric].columns)
                    metrics_dict[metric] = pd.concat([metrics_dict[metric], new_row.to_frame().T], ignore_index=True)
        return metrics_dict


class HyperparametersSelector:
    """
    A class for selecting hyperparameters based on given configurations.

    Attributes:
        hps (dict): A dictionary to store hyperparameter names and their selected values.

    Methods:
        Int(name, min_value, max_value, step): Returns an integer hyperparameter.
        Float(name, min_value, max_value, step): Returns a floating-point hyperparameter.
        Choice(name, values): Returns a choice from given possible values.
        _get_values_(): Returns the dictionary of all hyperparameter values.
    """

    def __init__(self, hps_configuration=None):
        """
        Initializes the HyperparametersSelection with an optional hyperparameters configuration.

        Args:
            hps_configuration (dict, optional): A pre-defined dictionary of hyperparameters. Defaults to None.
        """
        self.hps = {} if hps_configuration is None else hps_configuration

    def Int(self, name, min_value, max_value, step):
        """
        Generates or retrieves an integer hyperparameter within the specified range, using the specified step.

        Args:
            name (str): The name of the hyperparameter.
            min_value (int): The minimum value of the range.
            max_value (int): The maximum value of the range.
            step (int): The step between possible values within the range.

        Returns:
            int: The chosen or retrieved integer value for the hyperparameter.

        Raises:
            ValueError: If min_value is greater than max_value or if step is not positive.
        """
        if name in self.hps:
            return self.hps[name]
        if min_value > max_value:
            raise ValueError("min_val must be less than or equal to max_val")
        if step <= 0:
            raise ValueError("step must be a positive number")
        number = random.choice(range(min_value, max_value + 1, step))
        self.hps[name] = number
        return number

    def Float(self, name, min_value, max_value, step):
        """
        Generates or retrieves a floating-point hyperparameter within the specified range, using the specified step.

        Args:
            name (str): The name of the hyperparameter.
            min_value (float): The minimum value of the range.
            max_value (float): The maximum value of the range.
            step (float): The step between possible values within the range.

        Returns:
            float: The chosen or retrieved floating-point value for the hyperparameter.

        Raises:
            ValueError: If min_value is greater than max_value or if step is not positive.
        """
        if name in self.hps:
            return self.hps[name]
        if min_value > max_value:
            raise ValueError("min_val must be less than or equal to max_val")
        if step <= 0:
            raise ValueError("step must be a positive number")
        number = random.choice([min_value + i * step for i in range(int((max_value - min_value) / step) + 1)])
        self.hps[name] = number
        return number

    def Choice(self, name, values):
        """
        Chooses or retrieves a value from a list of possible values for a hyperparameter.

        Args:
            name (str): The name of the hyperparameter.
            values (list): A list of possible values from which to choose.

        Returns:
            Any: The chosen or retrieved value from the list.

        Raises:
            ValueError: If the list of possible values is empty.
        """
        if name in self.hps:
            return self.hps[name]
        if isinstance(values, list):
            if not values:
                raise ValueError("list cannot be empty")
        elif isinstance(values, np.ndarray):
            if values.size == 0:
                raise ValueError("array cannot be empty")
        election = random.choice(values)
        self.hps[name] = election
        return election

    def _get_values_(self):
        """
        Retrieves the dictionary containing all hyperparameter names and their selected values.

        Returns:
            dict: The dictionary of hyperparameter names and values.
        """
        return self.hps


class RandomSearch():
    """
    Clase RandomSearch para realizar una búsqueda aleatoria de hiperparámetros para un modelo de aprendizaje profundo.

    Atributos:
    ----------
    buildin_func : function
        Función que recibe un objeto de hiperparámetros y devuelve un modelo configurado.
    objective : str
        Métrica de validación que se utilizará para optimizar los hiperparámetros. Debe ser una de las siguientes: 
        ['val_loss', 'val_mae', 'val_mse', 'val_mape', 'val_r2'].
    max_trials : int
        Número máximo de configuraciones de hiperparámetros a probar.
    executions_per_trial : int
        Número de veces que se ejecutará el entrenamiento para cada configuración de hiperparámetros.
    results : dict
        Diccionario que almacena los resultados de las configuraciones probadas. Las keys son los números de los trials, 
        y los valores son una tupla con la configuración de hiperparámetros y el valor promedio de la métrica objetivo.

    Métodos:
    --------
    __init__(self, buildin_func, objective, max_trials, executions_per_trial):
        Inicializa la clase RandomSearch con la función de construcción de modelos, el objetivo a optimizar, 
        el número máximo de configuraciones (trials) y las ejecuciones por trial.

    search(self, X, Y, epochs, batch_size, validation_data, callbacks, verbose=1):
        Ejecuta la búsqueda de hiperparámetros. Entrena varios modelos con configuraciones de hiperparámetros aleatorios y
        selecciona el mejor basado en la métrica objetivo.
        
        Parámetros:
        -----------
        X : array-like
            Datos de entrada para el entrenamiento.
        Y : array-like
            Etiquetas o valores objetivo para el entrenamiento.
        epochs : int
            Número de épocas para entrenar cada modelo.
        batch_size : int
            Tamaño del lote utilizado durante el entrenamiento.
        validation_data : tuple
            Datos de validación en forma de (X_val, Y_val) para calcular la métrica objetivo.
        callbacks : list
            Lista de callbacks a utilizar durante el entrenamiento.
        verbose : int, opcional
            Nivel de verbosidad (1 por defecto). Determina qué tan detallada será la salida impresa.

    _generator_(self):
        Genera una nueva configuración de hiperparámetros utilizando la función de construcción y un objeto de selección de hiperparámetros.
        
        Retorna:
        --------
        hps_configuration : dict
            Diccionario que contiene los valores de la configuración de hiperparámetros generados.
        hps_object : object
            Objeto que contiene la información de los hiperparámetros seleccionados.
    """

    def __init__(self, building_func, objective, max_trials, executions_per_trial):
        self.building_func = building_func
        valid_objectives = ['val_loss', 'val_mae', 'val_mse', 'val_mape', 'val_r2']
        if objective not in valid_objectives:
            raise ValueError(f"'{objective}' is not a valid mode. Choose from {valid_objectives}.")
        self.objective = objective
        self.max_trials = max_trials
        self.executions_per_trial = executions_per_trial

    def search(self, X, Y, epochs, batch_size, validation_data, callbacks, verbose=1):
        """
        Realiza la búsqueda de hiperparámetros probando varias configuraciones de modelos.

        Parámetros:
        -----------
        X : array-like
            Conjunto de datos de entrada para el entrenamiento del modelo.
        Y : array-like
            Conjunto de etiquetas o valores objetivo para el entrenamiento.
        epochs : int
            Número de épocas para entrenar cada modelo.
        batch_size : int
            Tamaño del batch utilizado durante el entrenamiento.
        validation_data : tuple
            Conjunto de datos de validación en la forma (X_val, Y_val).
        callbacks : list
            Lista de callbacks a utilizar durante el entrenamiento.
        verbose : int, opcional
            Nivel de verbosidad del proceso de entrenamiento, por defecto es 1.

        Retorna:
        --------
        None
        """
        print('Searching: \n')
        results = {}
        for trial in range(self.max_trials):
            print(f'Searching model: {trial}')
            trial_results = []
            current_hps_configuration, current_hp_object = self._generator_()
            for model_trial in range(self.executions_per_trial):
                print(f'model trial: {model_trial}')
                model = self.building_func(current_hp_object)
                history = model.fit(x=X, y=Y, batch_size=batch_size, epochs=epochs, verbose=verbose, 
                                    callbacks=callbacks, validation_data=validation_data)

                # Seleccionar la métrica según el objetivo
                if self.objective == 'val_loss':
                    score = history.history['val_loss'][-1]
                    self.reverse = False
                elif self.objective == 'val_mae':
                    score = history.history['val_mae'][-1]
                    self.reverse = False
                elif self.objective == 'val_mse':
                    score = history.history['val_mse'][-1]
                    self.reverse = False
                elif self.objective == 'val_mape':
                    score = history.history['val_mape'][-1]
                    self.reverse = False
                elif self.objective == 'val_r2':
                    score = history.history['val_r2'][-1]
                    self.reverse = True

                trial_results.append(score)

            results[trial] = (current_hps_configuration, sum(trial_results) / len(trial_results))

        # Ordenar los resultados según la métrica objetivo
        results = {k: v for k, v in sorted(results.items(), key=lambda item: item[1][1], reverse=self.reverse)}
        results = {i: v for i, v in enumerate(results.values())}
        return results
    def _generator_(self):
        """
        Genera una nueva configuración de hiperparámetros aleatoria.

        Retorna:
        --------
        hps_configuration : dict
            Configuración generada de los hiperparámetros.
        hps_object : object
            Objeto que contiene los hiperparámetros seleccionados.
        """
        hps_object = HyperparametersSelector()
        _ = self.building_func(hps_object)
        hps_configuration = hps_object._get_values_()
        return hps_configuration, hps_object